# coding: utf-8

from __future__ import unicode_literals

import os

from lib2to3.pgen2 import token

from ebb_lint.checkers.registration import register_checker
from ebb_lint.errors import Errors


@register_checker("""

( import_from< 'from' pdb='pdb' any* >
| import_name< 'import' pdb='pdb' any* >
| dotted_as_name< pdb='pdb' any* >
)

""")
def check_for_pdb(pdb):
    yield pdb, Errors.no_debuggers, {}


@register_checker("""

power< any+ trailer< '.' func='set_trace' > trailer< '(' ')' > >

""")
def check_for_set_trace(func):
    yield func, Errors.no_debuggers, {}


@register_checker("""

f=file_input< any* >

""", comments_for=['f'])
def scan_top_level_comments(f, f_comments):
    if '# I sincerely swear that this is one-off code.' in f_comments:
        f.print_lint_ok = True
    return []


def scan_ancestry_for(node, attr, default):
    not_set = object()
    while node is not None:
        value = getattr(node, attr, not_set)
        if value is not not_set:
            return value
        node = node.parent
    return default


@register_checker("""

( simple_stmt< any* p='print' any* >
| print_stmt< p='print' any* >
| power< p='print' trailer< '(' any* ')' > any* >
)

""")
def check_for_print(p):
    if scan_ancestry_for(p, 'print_lint_ok', False):
        return
    yield p, Errors.no_print, {}


@register_checker(
    """

( import_from< 'from' mod=(NAME | dotted_name) any* >
| import_name< 'import' mod=(NAME | dotted_name) any* >
| dotted_as_name< mod=(NAME | dotted_name) any* >
)

    """,
    pass_filename=True, pass_future_features=True,
    python_disabled_version=(3, 0))
def check_for_implicit_relative_imports(
        filename, future_features, mod):  # ✘py33 ✘py34
    if 'absolute_import' in future_features:
        return
    [mod] = mod
    dirname = os.path.dirname(filename)
    segments = [l.value for l in mod.pre_order() if l.type == token.NAME]
    candidates = []
    candidates.extend(
        os.path.join(dirname,
                     *(segments[:-1] + ['{}.{}'.format(segments[-1], ext)]))
        for ext in ['py', 'pyc', 'pyo', 'pyd', 'so'])
    candidates.extend(
        os.path.join(dirname, *(segments + ['__init__.{}'.format(ext)]))
        for ext in ['py', 'pyc', 'pyo'])
    if not any(os.path.exists(candidate) for candidate in candidates):
        return

    node = next(mod.post_order())
    yield node, Errors.no_implicit_relative_imports, {}


@register_checker("""

( classdef< which='class' any* >
| funcdef< which='def' any* >
)

""", pass_filename=True)
def check_disallowed_dunder_init_statements(filename, which):
    if os.path.basename(filename) != '__init__.py':
        return
    yield which, Errors.no_definition_statements_in_dunder_init, {}


@register_checker("""

( suite< any* simple_stmt any* simple_stmt< p='pass' any > any* >
| if_stmt< 'if' any+ 'else' ':'
           suite< any* simple_stmt< p='pass' any > any* > >
| for_stmt< 'for' any+ 'else' ':'
            suite< any* simple_stmt< p='pass' any > any* > >
| while_stmt< 'while' any+ 'else' ':'
              suite< any* simple_stmt< p='pass' any > any* > >
| try_stmt< 'try' any+ 'finally' ':'
            suite< any* simple_stmt< p='pass' any > any* > >
)

""")
def check_useless_pass(p):
    yield p, Errors.useless_pass, {}


@register_checker("""

try_stmt< 'try' any+ 'except' ':'
          suite< any* simple_stmt< p='pass' any > any* >
          any* >

""")
def check_except_pass(p):
    yield p, Errors.no_except_pass, {}


@register_checker("""

( return_stmt< stmt='return' atom< lparen='(' any* ')' > >
| del_stmt< stmt='del' atom< lparen='(' any* ')' > >
| raise_stmt< stmt='raise' atom< lparen='(' any* ')' > >
| assert_stmt< stmt='assert' atom< lparen='(' any* ')' > >
| yield_expr< stmt='yield' atom< lparen='(' any* ')' > >
| print_stmt< stmt='print' atom< lparen='(' any* ')' > >
)

""")
def check_useless_parens(stmt, lparen):
    if lparen.prefix:
        return
    yield stmt, Errors.useless_parens, {'stmt': stmt.value}


@register_checker("""

yield_expr< stmt='yield' yield_arg< 'from' atom< lparen='(' any* ')' > > >

""", python_minimum_version=(3, 4))
def check_useless_parens_on_yield_from(stmt, lparen):  # ✘py27 ✘py33
    if lparen.prefix:
        return
    yield stmt, Errors.useless_parens, {'stmt': 'yield from'}


@register_checker("""

simple_stmt < power< f=('map' | 'filter') trailer< '(' any* ')' > any* > any* >

""")
def check_no_side_effects_function(f):
    [f] = f
    yield f, Errors.no_side_effects, {'thing': f.value}


_expr_type = {
    '[': 'a list comprehension',
    '{': 'a dict or set comprehension',
}


@register_checker("""

simple_stmt< ( atom< start='[' listmaker< any+ comp_for< any+ > > ']' >
             | atom< start='{' dictsetmaker< any+ comp_for< any+ > > '}' >
             ) any* >

""")
def check_no_side_effects_literal(start):
    yield start, Errors.no_side_effects, {'thing': _expr_type[start.value]}


@register_checker("""

power< f=('map' | 'filter') trailer< '(' arglist<
    lambdef ',' any*
> any* ')' > any* >

""")
def check_no_map_or_filter_with_lambda(f):
    [f] = f
    yield f, Errors.no_map_or_filter_with_lambda, {'func': f.value}


@register_checker("""

decorator< at='@' 'staticmethod' any* >

""")
def check_no_staticmethod_decorator(at):
    yield at, Errors.no_staticmethod_decorator, {}


# XXX: There's a bit of uncovered code below, but it's really just because I'm
# coding defensively. I don't know if it's possible to get lib2to3 to emit an
# AST that's in this particular shape, but I don't want to get caught offguard
# if it is. Maybe at some point I'll turn these into asserts or something?

class NoParentheizedGroup(Exception):
    pass


def parenthesized_group_leaves(container):
    if container is None or not container.children:  # pragma: nocover
        raise NoParentheizedGroup()
    first_child = container.children[0]
    last_child = container.children[-1]
    if not ((not first_child.children and first_child.value == '(')
            and (not last_child.children and last_child.value == ')')):
        raise NoParentheizedGroup()
    for child in container.children[1:-1]:
        for subchild in child.pre_order():
            if not subchild.children:
                yield subchild


@register_checker("""

atom=atom< first_string=STRING STRING+ >

""")
def check_for_unintentional_implicit_concatenation(atom, first_string):
    try:
        for child in parenthesized_group_leaves(atom.parent):
            if child.type != token.STRING:  # pragma: nocover
                break
        else:
            # all leaves were string literals
            return
    except NoParentheizedGroup:
        pass
    yield first_string, Errors.no_unintentional_implicit_concatenation, {}



# coding: utf-8

from __future__ import unicode_literals

import bisect
import io
import sys
from lib2to3.pgen2 import driver, token, tokenize
from lib2to3 import patcomp, pygram, pytree

import pep8
import six
import venusian
from intervaltree import Interval, IntervalTree

from ebb_lint._version import __version__
from ebb_lint.errors import Errors
from ebb_lint import checkers


_pep8_noqa = pep8.noqa
# This is a blight. Disable it unconditionally.
pep8.noqa = lambda ign: False


# detect_future_features isn't fully covered, but I don't really care, because
# I don't want to rewrite it. Maybe if it becomes more relevant I'll pull it
# out of this suite and actually properly unit test it, but right now I feel
# like it's mostly just working around a lib2to3 deficiency so I don't care
# enough to do anything else. It's stolen from lib2to3 directly. Why was this a
# private function? Ugh.

def detect_future_features(infile):  # pragma: nocover
    have_docstring = False
    gen = tokenize.generate_tokens(infile.readline)

    def advance():
        tok = next(gen)
        return tok[0], tok[1]

    ignore = frozenset((token.NEWLINE, tokenize.NL, token.COMMENT))
    features = set()
    try:
        while True:
            tp, value = advance()
            if tp in ignore:
                continue
            elif tp == token.STRING:
                if have_docstring:
                    break
                have_docstring = True
            elif tp == token.NAME and value == 'from':
                tp, value = advance()
                if tp != token.NAME or value != '__future__':
                    break
                tp, value = advance()
                if tp != token.NAME or value != 'import':
                    break
                tp, value = advance()
                if tp == token.OP and value == '(':
                    tp, value = advance()
                while tp == token.NAME:
                    features.add(value)
                    tp, value = advance()
                    if tp != token.OP or value != ',':
                        break
                    tp, value = advance()
            else:
                break
    except StopIteration:
        pass
    return frozenset(features)


if six.PY3:  # ✘py27
    def grammar_for_future_features(future_features):
        return pygram.python_grammar_no_print_statement

else:  # ✘py33 ✘py34
    def grammar_for_future_features(future_features):
        if 'print_function' in future_features:
            return pygram.python_grammar_no_print_statement
        else:
            return pygram.python_grammar


def find_comments(s, base_byte=0):
    fobj = io.StringIO(six.text_type(s))
    lines = Lines(fobj)
    fobj.seek(0)
    for typ, tok, spos, epos, _ in tokenize.generate_tokens(fobj.readline):
        if typ == tokenize.COMMENT:
            yield tok, Interval(
                lines.byte_of_pos(*spos) + base_byte,
                lines.byte_of_pos(*epos) + base_byte)


def read_file_using_source_encoding(filename):
    with open(filename, 'rb') as infile:
        encoding = tokenize.detect_encoding(infile.readline)[0]
    with io.open(filename, 'r', encoding=encoding) as infile:
        return infile.read()


def parse_source(driver, source):
    trailing_newline = not source or source.endswith('\n')
    # Thanks for this, lib2to3.
    if not trailing_newline:
        source += '\n'
    return driver.parse_string(source), trailing_newline


class Lines(object):
    def __init__(self, infile):
        count = 0
        self.lines = [(0, '')]
        for line in infile:
            self.lines.append((count, line))
            count += len(line)
        self.last_pos = len(self.lines) - 1, len(self.lines[-1][1])
        self.last_byte = count

    def __getitem__(self, idx):
        return self.lines[idx]

    def __iter__(self):
        for e, (count, line) in enumerate(self.lines):
            if e == 0:
                continue
            yield e, count, line

    def position_of_byte(self, byte):
        lineno = bisect.bisect_left(self.lines, (byte + 1,)) - 1
        column = byte - self.lines[lineno][0]
        return lineno, column

    def byte_of_pos(self, lineno, column):
        # This requires a bit of explanation. The source passed to lib2to3's
        # parser has an extra newline added in some cases, to deal with a bug
        # in lib2to3 where it crashes hard if files don't end with a trailing
        # newline. When that extra line is added, the final DEDENT token in the
        # file will have a lineno equal to the lines in the file plus one,
        # becase it's "at" a location that doesn't exist in the real file. If
        # this case wasn't specifically caught, the self[lineno] would raise an
        # exception because lineno is beyond the last index in self.lines. So,
        # when that case is detected, return the final byte position.
        if lineno == len(self.lines) and column == 0:
            return self.last_byte
        byte, _ = self[lineno]
        byte += column
        return byte

    def byte_of_node(self, node):
        return self.byte_of_pos(node.lineno, node.column)


def byte_intersection(tree, lower, upper):
    ret = 0
    for i in tree.search(lower, upper):
        ret += min(i.end, upper) - max(i.begin, lower)
    return ret


class EbbLint(object):
    name = 'ebb_lint'
    version = __version__

    collected_checkers = None
    _source = None
    _lines = None

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename
        self._intervals = {
            'comments': IntervalTree(),
            'string literals': IntervalTree(),
        }

    @classmethod
    def add_options(cls, parser):
        parser.add_option('--hard-max-line-length', default=119, type=int,
                          metavar='n',
                          help='absolute maximum line length allowed')
        parser.config_options.append('hard-max-line-length')
        parser.add_option('--permissive-bulkiness-percentage', default=67,
                          type=int, metavar='p', help=(
                              'integer percentage of a line which must be '
                              'string literals or comments to be allowed to '
                              'pass the soft line limit'))
        parser.config_options.append('permissive-bulkiness-percentage')

    @classmethod
    def parse_options(cls, options):
        # We implement our own line-length checker because it's not possible to
        # customize how another checker does its checking.
        options.ignore += 'E501',
        cls.options = options

        # This vastly speeds up the test suite, since parse_options is called
        # on every test now, and venusian does a lot of work.
        if cls.collected_checkers is not None:
            return

        collected_checkers = []

        def register_checker(pattern, checker, extra):
            if ('python_minimum_version' in extra
                    and sys.version_info < extra['python_minimum_version']):
                return
            if ('python_disabled_version' in extra
                    and sys.version_info > extra['python_disabled_version']):
                return
            pattern = patcomp.compile_pattern(pattern)
            collected_checkers.append((pattern, checker, extra))

        scanner = venusian.Scanner(register=register_checker)
        scanner.scan(checkers)
        cls.collected_checkers = collected_checkers

    @property
    def source(self):
        if self._source is None:
            self._source = read_file_using_source_encoding(self.filename)
        return self._source

    @property
    def lines(self):
        if self._lines is None:
            self._lines = Lines(self.source.splitlines(True))
        return self._lines

    def _message_for_node(self, node, error, **kw):
        line_offset = kw.pop('line_offset', None)
        if line_offset is None:
            byte = self.lines.byte_of_node(node) + kw.pop('offset', 0)
            lineno, column = self.lines.position_of_byte(byte)
        else:
            lineno = node.lineno + line_offset
            column = kw.pop('column')
        return self._message_for_pos((lineno, column), error, **kw)

    def _message_for_pos(self, pos, error, **kw):
        lineno, column = pos
        message = 'L{:03d} {}'.format(
            error.value.code, error.value.message.format(**kw))
        return lineno, column, message, type(self)

    def run(self):
        with open(self.filename, 'r') as infile:
            self.future_features = detect_future_features(infile)
        d = driver.Driver(
            grammar_for_future_features(self.future_features),
            convert=pytree.convert)
        tree, trailing_newline = parse_source(d, self.source)
        if not trailing_newline:
            yield self._message_for_pos(
                self.lines.last_pos, Errors.no_trailing_newline)

        for error in self._check_tree(tree):
            yield error

        for error in self._check_line_lengths():
            yield error

    def _check_tree(self, tree):
        for node in tree.pre_order():
            for error in self._scan_node_for_ranges(node):
                yield error

            for pattern, checker, extra in self.collected_checkers:
                results = {}
                if not pattern.match(node, results):
                    continue
                for k in extra.get('comments_for', ()):
                    results[k + '_comments'] = [
                        c for c, _ in find_comments(node.prefix)]
                if extra.get('pass_filename', False):
                    results['filename'] = self.filename
                if extra.get('pass_future_features', False):
                    results['future_features'] = self.future_features
                for error_node, error, kw in checker(**results):
                    yield self._message_for_node(error_node, error, **kw)

    def _scan_node_for_ranges(self, node):
        if node.children or (node.type != token.STRING and not node.prefix):
            return

        byte = self.lines.byte_of_node(node)

        if node.type == token.STRING:
            self._intervals['string literals'].add(Interval(
                byte, byte + len(node.value)))

        comments = list(
            find_comments(node.prefix, byte - len(node.prefix)))
        for c, i in comments:
            self._intervals['comments'].add(i)
            m = _pep8_noqa(c)
            if m is not None:
                yield self._message_for_pos(
                    self.lines.position_of_byte(i.begin + m.start()),
                    Errors.no_noqa)

    def _check_line_lengths(self):
        soft_limit = self.options.max_line_length
        hard_limit = self.options.hard_max_line_length
        permitted_percentage = self.options.permissive_bulkiness_percentage
        for lineno, line_start, line in self.lines:
            line = line.rstrip('\r\n')
            if len(line) <= soft_limit:
                continue
            if len(line) > hard_limit:
                yield self._message_for_pos(
                    (lineno, hard_limit), Errors.line_too_long,
                    length=len(line), which_limit='hard', limit=hard_limit,
                    extra='')
                continue

            line_end = line_start + len(line)
            percentages = {}
            for name, i in self._intervals.items():
                n_bytes = byte_intersection(i, line_start, line_end)
                percentages[name] = p = n_bytes * 100 // len(line)
                assert 0 <= p <= 100, 'line percentage not in range'

            if any(p >= permitted_percentage for p in percentages.values()):
                continue

            extra = ' since the line has ' + '; '.join(
                '{p}% {name}'.format(p=p, name=name)
                for name, p in percentages.items())
            yield self._message_for_pos(
                (lineno, soft_limit), Errors.line_too_long,
                length=len(line), which_limit='soft', limit=soft_limit,
                extra=extra)



from __future__ import unicode_literals

import ast
import functools
import re
import sys

import pytest
import six
from flake8.engine import get_parser

from ebb_lint.flake8 import EbbLint, Lines


py2skip = pytest.mark.skipif(not six.PY3, reason='not runnable on python 2')
py3skip = pytest.mark.skipif(six.PY3, reason='not runnable on python 3')
pre_py34skip = pytest.mark.skipif(
    sys.version_info < (3, 4), reason='not runnable before python 3.4')


_code_pattern = re.compile(r"""
    (?x)
    \$([A-Z]\d{3})\$
""")


def find_error_locations(source):
    error_locations = []

    def replacement(match):
        last_offset = error_locations[-1][2] if error_locations else 0
        new_offset = last_offset + match.end() - match.start()
        error_locations.append((
            match.start() - last_offset, match.group(1), new_offset))
        return ''

    source = _code_pattern.sub(replacement, source)
    lines = Lines(source.splitlines(True))
    return source, [
        lines.position_of_byte(b) + (code,) for b, code, _ in error_locations]


@pytest.fixture(autouse=True)
def scan_for_checkers(request):
    parser, options_hooks = get_parser()
    args_marker = request.node.get_marker('flake8_args')
    if args_marker is None:
        args = []
    else:
        args = list(args_marker.args)
    opts, args = parser.parse_args(args)
    opts.ignore = tuple(opts.ignore)
    EbbLint.parse_options(opts)


all_sources = [
    '',

    """

def f():
    $L102$'''
    spam
    '''

    """,

    """

def f():
    $L102$'''spam'''

    """,

    """

def f():
    $L102$'spam'

    """,

    """

def f():
    $L102$"spam"

    """,

    '''

def f():
    """$L102$ spam"""

    ''',

    '''

def f():
    """spam$L102$ """

    ''',

    '''

def __init__():
    $L101$"""
    spam
    """

    ''',

    '''

class Spam(object):
    """
    It's spam.
    """

    ''',

    '''

import $L203$pdb
pdb.$L203$set_trace()

    ''',

    '''

import $L203$pdb as not_pdb
not_pdb.$L203$set_trace()

    ''',

    '''

from $L203$pdb import set_trace
set_trace()

    ''',

    '''

import pudb
pudb.$L203$set_trace()

    ''',

    '''

x = $L204$'a' 'b'

    ''',

    '''

x = [
    $L204$'a'
    'b',
    'c',
]

    ''',

    '''

x = [
    ($L204$'a'
     'b'.format()),
    'c',
]

    ''',

    '''

x = [
    ($L204$'a'
     'b'()),
    'c',
]

    ''',

    '''

x = [
    [$L204$'a'
     'b'],
    'c',
]

    ''',

    '''

x = [
    ('a'
     'b').format(),
    'c',
]

    ''',

    '''

x = [
    ('a'
     'b'),
    'c',
]

    ''',

    '''

x = ('a' 'b')

    ''',

    '''

from __future__ import print_function
import sys

$L202$print("Yay!", file=sys.stdout)

    ''',

    '''

# I sincerely swear that this is one-off code.
from __future__ import print_function
import sys

print("Yay!", file=sys.stdout)

    ''',

    '''

def f():
    pass$L301$''',

    '''

# coding: utf-8
x = '\ufffd' + $L204$'\ufffe' '\uffff'

    ''',

    '''

def f():
    """
    Spam; eggs.
    """
    $L207$pass

    ''',

    '''

if True:
    pass
else:
    $L207$pass

    ''',

    '''

for x in y:
    pass
else:
    $L207$pass

    ''',

    '''

while True:
    pass
else:
    $L207$pass

    ''',

    '''

try:
    pass
finally:
    $L207$pass

    ''',

    '''

try:
    pass
except Whatever:
    do_something()
    $L207$pass

    ''',

    '''

try:
    pass
except:
    $L208$pass

    ''',

    '''

try:
    pass
except Whatever:
    pass

    ''',

    '''

try:
    pass
except Whatever:
    pass
except:
    $L208$pass

    ''',

    '''

try:
    pass
except:
    $L208$pass
except Whatever:
    pass

    ''',

    '''

try:
    pass
except:
    $L208$pass
else:
    do_something()
finally:
    do_something_else()

    ''',

    '''

class C(object):
    $L212$@staticmethod
    def f():
        pass

    ''',

    '''

class C(object):
    @classmethod
    def f(cls):
        pass

    ''',

    '''

class C(object):
    f = staticmethod(g)

    ''',

    '''

spam  $L303$# noqa

    ''',

    '''

spam  # eggs $L303$# noqa

    ''',

    '''

$L204$'a' 'b'  $L303$# noqa

    ''',
]


line_length_test_mark_generator = functools.partial(
    pytest.mark.flake8_args,
    '--max-line-length', '15', '--hard-max-line-length', '25')


line_length_test_mark = line_length_test_mark_generator()
all_sources.extend(line_length_test_mark(s) for s in [
    '''

AbcdeAbcdeAbcde

    ''',

    '''

AbcdeAbcdeAbcde$L302$f

    ''',

    '''

AbcdeAbcdeAbcd $L302$= 'e'

    ''',

    '''

AbcdeAbcdeAbcd $L302$= """

This is a long
docstring. It's
not wide, but
it has more
than 25
characters.

"""

    ''',

    '''

A = """

This is a long
docstring. It's
not wide, but
it has more
than 25
characters.

""", AbcdeAbcde$L302$Abcde

    ''',

    '''

A = 'bcdeAbcdeAbcde'

    ''',

    '''

A = 'bcdeAbcd' $L302$ # eA

    ''',

    '''

AbcdeAbcdeAbcdeAbcdeAbcde$L302$f

    ''',

    '''

AbcdeAbcdeAbcdeAbcdeAbcde$L302$fghijklmnopqrstuvwxyz

    ''',
])


def permissive_percentage_test_mark(p, s):
    return line_length_test_mark_generator(
        '--permissive-bulkiness-percentage', str(p))(s)


all_sources.extend(permissive_percentage_test_mark(p, s) for p, s in [
    (100, '''

"AbcdeAbcdeAbcdeAbc"

    '''),

    (100, '''

# AbcdeAbcdeAbcdeAbc

    '''),

    (100, '''

a = "AbcdeAbcde$L302$Abcde"

    '''),

    (100, '''

a # AbcdeAbcdeA$L302$bcdeA

    '''),

    (50, '''

A = 'bcdeAbcd'  # eA

    '''),

    (50, '''

A = 'bcde'  # A$L302$bcdeA

    '''),

    (20, '''

A = 'bcdeAbcd'  # eA

    '''),

    (0, '''

A = 'bcdeAbcd'  # eA

    '''),
])


hard_limit_only_test_mark = pytest.mark.flake8_args(
    '--max-line-length', '15', '--hard-max-line-length', '15')
all_sources.extend(hard_limit_only_test_mark(s) for s in [
    '''

AbcdeAbcdeAbcde

    ''',

    '''

AbcdeAbcdeAbcde$L302$f

    ''',

    '''

AbcdeAbcdeAbcd $L302$= 'e'

    ''',

    '''

A = 'bcdeAbcdeA$L302$bcde'

    ''',

    '''

A = 'bcdeAbcd' $L302$ # eA

    ''',

    '''

AbcdeAbcdeAbcde$L302$AbcdeAbcdef

    ''',

    '''

AbcdeAbcdeAbcde$L302$AbcdeAbcdefghijklmnopqrstuvwxyz

    ''',
])


contexts = [
    ('# I sincerely swear that this is one-off code.', ''),

    ("""
# I sincerely swear that this is one-off code.
# I sincerely swear that I am not a member of the Communist Party.
    """, ''),

    ('', '$L202$'),
]


all_sources.extend(
    py3skip('''
{ctx[0]}
{ctx[1]}{print_}
    '''.format(ctx=ctx, print_=print_))
    for ctx in contexts
    for print_ in [
        "print 'hi'",
        "print 'hi',",
        "print >> aether, 'hi'",
        "print",
    ])


all_sources.extend(
    '''
{ctx[0]}
from __future__ import print_function
{ctx[1]}{print_}
    '''.format(ctx=ctx, print_=print_)
    for ctx in contexts
    for print_ in [
        "print('hi')",
        "print()",
    ])


element_pairs = [
    ('1', '2'),
    ("'spam'", "'eggs'"),
]


delimiter_pairs = [
    (('(', ')'), ['f(x)']),
    (('[', ']'), ['f(x)']),
    (('{', '}'), ['f(x)', 'f(x): g(x)']),
]


all_sources.extend(
    template.format(delim=delim, elem=elem)
    for delim, _ in delimiter_pairs
    for elem in element_pairs
    for template in [
        '''

spam = {delim[0]}
    {elem[0]},
    {elem[1]},
{delim[1]}

        ''',

        '''

spam = {delim[0]}{elem[0]}, {elem[1]}, """

"""{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {elem[0]},
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {elem[0]},
    $L201${elem[1]}
{delim[1]}

        ''',

        '''

spam = {delim[0]}{elem[0]}, {elem[1]}{delim[1]}

        ''',

        '''

spam = {delim[0]}{elem[0]}{delim[1]}

        ''',
    ])


all_sources.extend(
    template.format(delim=delim, elem=elem)
    for delim, _ in delimiter_pairs
    for elem in element_pairs
    for template in [
        '''

spam = {delim[0]}
    $L201${elem[0]}
{delim[1]}

        ''',

        '''

$L210${delim[0]}{elem[0]} for x in y{delim[1]}

        ''',

        '''

a = {delim[0]}{elem[0]} for x in y{delim[1]}

        ''',
    ]
    # Ignore this because it's a parenthesized expression, and not a tuple.
    if delim[0] != '(')


all_sources.extend(
    template.format(delim=delim)
    for delim, _ in delimiter_pairs
    for template in [
        '''

spam = {delim[0]}{delim[1]}

        ''',

        '''

spam = {delim[0]}
{delim[1]}

        ''',
    ])


all_sources.extend(
    template.format(delim=delim, expr=expr)
    for delim, exprs in delimiter_pairs
    for expr in exprs
    for template in [
        '''

spam = {delim[0]}
    {expr} for x in y
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr}
    for x in y
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr}
    for x in y
    if p(x)
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr}
    for y in z
    for x in y
    if p(x)
{delim[1]}

        ''',

        '''

spam = {delim[0]}
    {expr} for x in y{delim[1]}

        ''',

        '''

spam = {delim[0]}{expr} for x in y{delim[1]}

        ''',
    ])


dict_element_pairs = [
    (('1', '2'), ('3', '4')),
    (("'spam'", "'eggs'"), ("'eggs'", "'spam'")),
]


all_sources.extend(
    template.format(elem=elem)
    for elem in dict_element_pairs
    for template in [
        '''

spam = {{
    {elem[0][0]}: {elem[0][1]},
    {elem[1][0]}: {elem[1][1]},
}}

        ''',

        '''

spam = {{
    {elem[0][0]}: {elem[0][1]},
    {elem[1][0]}: $L201${elem[1][1]}
}}

        ''',

        '''

spam = {{{elem[0][0]}: {elem[0][1]}, {elem[1][0]}: {elem[1][1]}}}

        ''',
    ])


no_paren_statements = ['del', 'assert', 'return', 'raise', 'yield']


all_sources.extend(
    template.format(stmt=stmt)
    for stmt in no_paren_statements
    for template in [
        '''

$L209${stmt}(a)

        ''',

        '''

{stmt} (a)

        ''',
    ])

all_sources.extend([
    '''

a = $L209$yield(b)

    ''',

    '''

a = yield (b)

    ''',
])

all_sources.extend(pre_py34skip(s) for s in [
    '''

a = $L209$yield from(b)

    ''',

    '''

a = yield from (b)

    ''',

    '''

$L209$yield from(b)

    ''',

    '''

yield from (b)

    ''',
])

all_sources.extend(py3skip(s) for s in [
    '''

$L202$$L209$print(b)

    ''',

    '''

$L202$print (b)

    ''',
])


all_sources.extend(
    template.format(func=func)
    for func in ['map', 'filter']
    for template in [
        '''

$L210${func}(x, y)

        ''',

        '''

a = {func}(x, y)

        ''',

        '''

a = $L211${func}(lambda x: x, y)

        ''',

        '''

a = $L211${func}(lambda x1, x2: x1 + x2, y, z)

        ''',
    ])


docstrings = [
    """
    hi
    """,

    """
    hi

    hello
    """,

    """
    hi

      hello

    hey
    """,

    """
        hi
        """,

    """
        hi

        hello
        """,

    """
        hi

          hello

        hey
        """,

    '\nhi\n',

    '\nhi\n\nhello\n',

    '\nhi\n\n  hello\n\nhey\n',

    """
    hi

$L102$  hey

    hello
    """,

    """
$L102$
    hi
    """,

    """
    $L102$
    hi
    """,

    '\n    hi\n$L102$',

    """
$L102$   hey
   """,

    """
$L102$   hey
   $L102$ """,

    'hi',
]


sphinx_fields = [
    'param', 'parameter', 'arg', 'argument', 'key', 'keyword', 'type',
    'raises', 'raise', 'except', 'exception', 'var', 'ivar', 'cvar',
    'vartype', 'returns', 'return', 'rtype',
]


docstrings.extend("""
    $L104$:{} spam: eggs
    """.format(field) for field in sphinx_fields)


docstrings.extend("""
    $L104$:{}: spam
    """.format(field) for field in sphinx_fields)


all_sources.extend('''

def f():
    """{}"""

'''.format(docstring) for docstring in docstrings)


all_sources.extend('''

def f():
    $L102$"""{}"""

'''.format(docstring) for docstring in [
    """
""",

    """
 """,

    """
hi""",

    """hi
    """,

    '',
])


test_docstrings = [
    """
    $L103$Test that the thing does the thing.
    """,

    '$L103$Test that the thing does the thing.',

    """
    $L103$VERIFY MY YELLING VOLUME
    """,

    """
    MY YELLING VOLUME IS FINE
    """,
]


test_docstrings.extend("""
    $L103${} a thing.
    """.format(word) for word in [
    'test', 'verify', 'ensure', 'tests', 'verifies', 'ensures', 'testing',
    'verifying', 'ensuring',
])


all_sources.extend('''

def test_f():
    """{}"""

'''.format(docstring) for docstring in test_docstrings)


dunder_init_sources = [
    '''

$L205$def spam():
    pass

    ''',

    '''

$L205$class Spam():
    pass

    ''',

    '''

from spam import eggs

__all__ = ['eggs']

    ''',
]


all_filename_sources = [
    ('__init__.py', source) for source in dunder_init_sources]


@pytest.fixture
def assert_lint():
    def check(source, sourcefile, no_errors=False):
        # Strip trailing spaces, since the most comfortable triple-quoted
        # string format will have extra spaces on the last line. If this isn't
        # removed, it'll throw off the trailing newline checker.
        source = source.rstrip(' ')
        source, error_locations = find_error_locations(source)
        if no_errors:
            error_locations = []
        sourcefile.write_text(source, encoding='utf-8')
        lint = EbbLint(ast.parse(source), sourcefile.strpath)
        actual = [
            (line, col, message[:4]) for line, col, message, _ in lint.run()]
        assert actual == error_locations

    return check


@pytest.mark.parametrize(('filename', 'source'), all_filename_sources)
def test_linting_with_filename(assert_lint, tmpdir, source, filename):
    sourcefile = tmpdir.join(filename)
    assert_lint(source, sourcefile)


@pytest.mark.parametrize('source', all_sources)
def test_linting_with_default_filename(assert_lint, tmpdir, source):
    sourcefile = tmpdir.join('source.py')
    assert_lint(source, sourcefile)


implicit_relative_import_sources = [
    {
        'spam.py': '''

import $L206$eggs

        ''',

        'eggs.py': '''

import $L206$spam

        ''',
    },

    {
        'spam.py': '''

import eggs

        ''',
    },

    {
        'spam.py': '''

import $L206$eggs.eggs

        ''',

        'eggs/__init__.py': '',
        'eggs/eggs.py': '',
    },

    {
        'spam.py': '''

from __future__ import absolute_import

import eggs

        ''',

        'eggs.py': '''

from __future__ import absolute_import

import spam

        ''',
    },

    {
        'spam.py': '''

import eggs.eggs

        ''',
    },
]


implicit_relative_import_sources.extend(
    {
        'spam.py': '''

import $L206$eggs

        ''',

        'eggs.{}'.format(ext): '',
    }
    for ext in ['py', 'pyc', 'pyo', 'pyd', 'so']
)


implicit_relative_import_sources.extend(
    {
        'spam.py': '''

import $L206$eggs

        ''',

        'eggs/__init__.{}'.format(ext): '',
    }
    for ext in ['py', 'pyc', 'pyo']
)


all_implicit_relative_import_sources = [
    (name, sources)
    for sources in implicit_relative_import_sources
    for name in sources]


@pytest.mark.parametrize(
    ('to_test', 'sources'), all_implicit_relative_import_sources)
@py3skip
def test_linting_implicit_relative_imports(
        assert_lint, tmpdir, to_test, sources):
    for name, source in sources.items():
        tmpdir.join(name).write_text(source, encoding='utf-8', ensure=True)
    assert_lint(sources[to_test], tmpdir.join(to_test))


@pytest.mark.parametrize(
    ('to_test', 'sources'), all_implicit_relative_import_sources)
@py2skip
def test_linting_implicit_relative_imports_disabled_on_py3(
        assert_lint, tmpdir, to_test, sources):
    for name, source in sources.items():
        tmpdir.join(name).write_text(source, encoding='utf-8', ensure=True)
    assert_lint(sources[to_test], tmpdir.join(to_test), no_errors=True)
